from datetime import date, datetime, time
from typing import Optional, Type

from _decimal import Decimal

from qtgql.codegen.py.scalars import BaseCustomScalar


class DateTimeScalar(BaseCustomScalar[datetime]):
    """An ISO-8601 encoded datetime."""

    GRAPHQL_NAME: str = "DateTime"
    DEFAULT_VALUE = datetime.now()

    @classmethod
    def from_graphql(cls, v: Optional[str] = None) -> "DateTimeScalar":
        if v:
            return cls(datetime.fromisoformat(v))
        return cls()

    def to_qt(self) -> str:
        return self._value.strftime("%H:%M (%m/%d/%Y)")


class DateScalar(BaseCustomScalar[date]):
    """An ISO-8601 encoded date."""

    GRAPHQL_NAME = "Date"
    DEFAULT_VALUE = date(year=1998, month=8, day=23)

    @classmethod
    def from_graphql(cls, v: Optional[str] = None) -> "DateScalar":
        if v:
            return cls(date.fromisoformat(v))
        return cls()

    def to_qt(self) -> str:
        return self._value.isoformat()


class TimeScalar(BaseCustomScalar[time]):
    """an ISO-8601 encoded time."""

    GRAPHQL_NAME = "Time"
    DEFAULT_VALUE = time()

    @classmethod
    def from_graphql(cls, v: Optional[str] = None) -> "TimeScalar":
        if v:
            return cls(time.fromisoformat(v))
        return cls()

    def to_qt(self) -> str:
        return self._value.isoformat()


class DecimalScalar(BaseCustomScalar[Decimal]):
    """A Decimal value serialized as a string."""

    GRAPHQL_NAME = "Decimal"
    DEFAULT_VALUE = Decimal()

    @classmethod
    def from_graphql(cls, v: Optional[str] = None) -> "DecimalScalar":
        if v:
            return cls(Decimal(v))
        return cls()

    def to_qt(self) -> str:
        return str(self._value)


CustomScalarMap = dict[str, Type[BaseCustomScalar]]
CUSTOM_SCALARS: CustomScalarMap = {
    DateTimeScalar.GRAPHQL_NAME: DateTimeScalar,
    DecimalScalar.GRAPHQL_NAME: DecimalScalar,
    DateScalar.GRAPHQL_NAME: DateScalar,
    TimeScalar.GRAPHQL_NAME: TimeScalar,
}



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Any

from attrs import define
from jinja2 import Environment, PackageLoader, select_autoescape

if TYPE_CHECKING:  # pragma: no cover
    from qtgql.codegen.py.config import QtGqlConfig
    from qtgql.codegen.py.objecttype import GqlEnumDefinition, GqlTypeDefinition

env = Environment(loader=PackageLoader("qtgql.codegen.py"), autoescape=select_autoescape())

SchemaTemplate = env.get_template("schema.jinja.py")


@define
class TemplateContext:
    enums: list[GqlEnumDefinition]
    types: list[GqlTypeDefinition]
    config: QtGqlConfig

    @property
    def dependencies(self) -> list[str]:
        def build_import_statement(t: type[Any]) -> str:
            mod = inspect.getmodule(t)
            assert mod
            return f"from {mod.__name__} import {t.__name__}"

        ret = [build_import_statement(scalar) for scalar in self.config.custom_scalars.values()]
        ret.append(build_import_statement(self.config.base_object))
        return ret

    @property
    def custom_scalars(self) -> list[str]:
        return [scalar.__name__ for scalar in self.config.custom_scalars.values()]

    @property
    def base_object_name(self) -> str:
        return self.config.base_object.__name__


def py_template(context: TemplateContext) -> str:
    return SchemaTemplate.render(context=context)



from __future__ import annotations

from typing import Generic, Optional, TypeVar

from PySide6.QtCore import QAbstractListModel, QByteArray, QObject, Qt, Signal
from typing_extensions import Self

from qtgql import qproperty, slot

__all__ = ["QGraphQListModel", "get_base_graphql_object"]


class _BaseQGraphQLObject(QObject):
    type_map: dict[str, type[_BaseQGraphQLObject]]

    __singleton__: _BaseQGraphQLObject

    def __init_subclass__(cls, **kwargs):
        cls.type_map[cls.__name__] = cls

    def __init__(self, parent: Optional[QObject] = None):
        super().__init__(parent)

    @classmethod
    def default_instance(cls) -> Self:  # type: ignore
        try:
            return cls.__singleton__
        except AttributeError:
            cls.__singleton__ = cls()
            return cls.__singleton__

    @classmethod
    def from_dict(
        cls, parent: T_BaseQGraphQLObject, data: dict
    ) -> T_BaseQGraphQLObject:  # pragma: no cover
        raise NotImplementedError


T_BaseQGraphQLObject = TypeVar("T_BaseQGraphQLObject", bound=_BaseQGraphQLObject)


class QGraphQListModel(QAbstractListModel, Generic[T_BaseQGraphQLObject]):
    OBJECT_ROLE = Qt.ItemDataRole.UserRole + 1
    _role_names = {OBJECT_ROLE: QByteArray("object")}  # type: ignore
    currentIndexChanged = Signal()

    def __init__(
        self,
        parent: Optional[QObject],
        default_object: T_BaseQGraphQLObject,
        data: list[T_BaseQGraphQLObject],
    ):
        super().__init__(parent)

        self._data = data
        self._default_object = default_object
        self._current_index: int = 0

    @slot
    def set_current_index(self, i: int) -> None:
        self._current_index = i
        self.currentIndexChanged.emit()

    @qproperty(int, notify=currentIndexChanged, fset=set_current_index)
    def currentIndex(self) -> int:
        return self._current_index

    @qproperty(QObject, notify=currentIndexChanged)  # type: ignore
    def currentObject(self) -> Optional[T_BaseQGraphQLObject]:
        try:
            return self._data[self._current_index]
        except IndexError:
            return self._default_object

    def rowCount(self, *args, **kwargs) -> int:
        return len(self._data)

    def roleNames(self) -> dict:
        return self._role_names  # type: ignore

    def data(self, index, role=...) -> Optional[T_BaseQGraphQLObject]:
        if index.row() < len(self._data) and index.isValid():
            if role == self.OBJECT_ROLE:
                return self._data[index.row()]
            raise NotImplementedError(
                f"role {role} is not a valid role for {self.__class__.__name__}"
            )

    def append(self, node: T_BaseQGraphQLObject) -> None:
        count = self.rowCount()
        self.beginInsertRows(self.index(count), count, count)
        self._data.append(node)
        self.endInsertRows()

    @slot
    def pop(self, index: Optional[int] = None) -> None:
        index = -1 if index is None else index
        real_index = index if index > -1 else self.rowCount()
        self.beginRemoveRows(self.index(index - 1).parent(), real_index, real_index)
        self._data.pop(index)
        self.endRemoveRows()


def get_base_graphql_object(name: str) -> type[_BaseQGraphQLObject]:
    """
    :param name: valid attribute name (used by codegen to import it).
    :returns: A type to be extended by all generated types.
    """
    type_map: dict[str, type[_BaseQGraphQLObject]] = {}
    return type(name, (_BaseQGraphQLObject,), {"type_map": type_map})  # type: ignore


BaseGraphQLObject = get_base_graphql_object("BaseGraphQLObject")

T_BaseModel = TypeVar("T_BaseModel", bound=QGraphQListModel)



from pathlib import Path
from typing import Callable, Type

import requests  # type: ignore
from attrs import define

from qtgql.codegen.introspection import SchemaEvaluator, introspection_query
from qtgql.codegen.py.bases import BaseGraphQLObject, _BaseQGraphQLObject
from qtgql.codegen.py.compiler import TemplateContext, py_template
from qtgql.codegen.py.custom_scalars import CUSTOM_SCALARS, CustomScalarMap


@define
class QtGqlConfig:
    """Encapsulates configurations for a qtgql-codegen application.

    :arg url:
    """

    url: str
    output: Path
    """full path for the generated output file."""
    evaluator: Type[SchemaEvaluator] = SchemaEvaluator
    """evaluates the schema and generates types."""
    custom_scalars: CustomScalarMap = CUSTOM_SCALARS
    """mapping of custom scalars, respected by the schema evaluator."""
    template_class: Callable[[TemplateContext], str] = py_template
    """jinja template."""
    base_object: Type[_BaseQGraphQLObject] = BaseGraphQLObject
    """base object to be extended by all generated types."""
    qml_import_name: str = "QtGql"
    """Enums would exist in this namespace."""

    def fetch(self) -> None:
        res = requests.post(self.url, json={"query": introspection_query})
        introspected = res.json()["data"]
        self.evaluator.from_dict(introspected, config=self).dump(self.output)

    def __attrs_post_init__(self):
        if self.custom_scalars != CUSTOM_SCALARS:
            self.custom_scalars.update(CUSTOM_SCALARS)



from __future__ import annotations

import contextlib
import enum
from functools import cached_property
from typing import Any, Optional, Type

from attrs import define

from qtgql.codegen.py.bases import QGraphQListModel, _BaseQGraphQLObject
from qtgql.codegen.py.custom_scalars import CustomScalarMap
from qtgql.codegen.py.scalars import BaseCustomScalar, BuiltinScalar
from qtgql.codegen.utils import AntiForwardRef
from qtgql.typingref import TypeHinter


class Kinds(enum.Enum):
    SCALAR = "SCALAR"
    OBJECT = "OBJECT"
    ENUM = "ENUM"
    LIST = "LIST"
    UNION = "UNION"
    NON_NULL = "NON_NULL"
    INTERFACE = "INTERFACE"
    INPUT_OBJECT = "INPUT_OBJECT"


@define(slots=False)
class FieldProperty:
    name: str
    type: GqlTypeHinter
    type_map: dict[str, GqlTypeDefinition]
    enums: "EnumMap"
    scalars: CustomScalarMap
    description: Optional[str] = ""

    @cached_property
    def default_value(self):
        if builtin_scalar := self.type.is_builtin_scalar:
            if builtin_scalar.tp is str:
                return f"'{builtin_scalar.default_value}'"
            return f"{builtin_scalar.default_value}"
        if object_def := self.type.is_object_type:
            return f"{object_def.name}(parent=self)"

        if model_of := self.type.is_model:
            return (
                f"{QGraphQListModel.__name__}(parent=self, data=[], "
                f"default_object={model_of.name}.{_BaseQGraphQLObject.default_instance.__name__}())"
            )

        if custom_scalar := self.type.is_custom_scalar(self.scalars):
            return f"SCALARS.{custom_scalar.__name__}()"

        if enum_def := self.type.is_enum:
            return f"{enum_def.name}(1)"  # 1 is where auto() starts.

        return "None"  # Unions are not supported yet.

    @cached_property
    def deserializer(self) -> str:
        """This gets the dict from graphql and passes the data to init, goes to
        `from_graphql` on the J2 template
        The J2 template in this context provides the data for the field (if was existed
        in the dict) using walrus operator (attribute name is the field name).
        ."""
        # every thing is possibly optional since you can query for only so or so fields.

        if self.type.is_builtin_scalar:
            return self.name

        if scalar := self.type.is_custom_scalar(self.scalars):
            return (
                f"SCALARS.{scalar.__name__}.{BaseCustomScalar.from_graphql.__name__}({self.name})"
            )
        if model_of := self.type.is_model:
            return (
                f"{QGraphQListModel.__name__}("
                f"parent=parent, "
                f"data=[{model_of.name}.from_dict(parent, data) for data in {self.name}], "
                f"default_object={model_of.name}.{_BaseQGraphQLObject.default_instance.__name__}()"
                f")"
            )
        if self.type.is_union():
            return f"cls.type_map[{self.name}['__typename']].from_dict(parent, {self.name})"
        if gql_type := self.type.is_object_type:
            return f"{gql_type.name}.from_dict(parent, {self.name})"
        if enum_def := self.type.is_enum:
            return f"{enum_def.name}[data.get('{self.name}', {self.default_value})]"  # graphql enums evaluates to string of the name.
        raise NotImplementedError  # pragma: no cover

    @cached_property
    def annotation(self) -> str:
        """
        :returns: Annotation of the field based on the real type,
        meaning that the private attribute would be of that type.
        this goes for init and the property setter.
        """
        return self.type.annotation(self.scalars)

    @cached_property
    def fget_annotation(self) -> str:
        """This annotates the value that is QML-compatible."""
        if custom_scalar := self.type.is_custom_scalar(self.scalars):
            return TypeHinter.from_annotations(
                custom_scalar.to_qt.__annotations__["return"]
            ).stringify()
        if self.type.is_enum:
            return "int"

        return self.annotation

    @cached_property
    def property_type(self) -> str:
        try:
            # this should raise if it is an inner type.
            ret = GqlTypeHinter.from_string(self.fget_annotation, self.type_map)
            if ret.type in self.type_map.values():
                raise TypeError  # in py3.11 the get_type_hints won't raise so raise brutally
            return self.fget_annotation
        except (TypeError, NameError):
            if self.type.is_union():
                return "QObject"  # graphql doesn't support scalars in Unions ATM. (what about Enums in unions)?
            if self.type.is_enum:
                # QEnum value must be int
                return "int"
            # might be a model, which is also QObject
            assert self.type.is_model or self.type.is_object_type
            return "QObject"

    @cached_property
    def setter_name(self) -> str:
        return self.name + "_setter"

    @cached_property
    def signal_name(self) -> str:
        return self.name + "Changed"

    @cached_property
    def private_name(self) -> str:
        return "_" + self.name

    @property
    def fget(self) -> str:
        if self.type.is_custom_scalar(self.scalars):
            return f"return self.{self.private_name}.{BaseCustomScalar.to_qt.__name__}()"
        if self.type.is_enum:
            return f"return self.{self.private_name}.value"
        return f"return self.{self.private_name}"


@define(slots=False)
class GqlTypeDefinition:
    kind: Kinds
    name: str
    fields: list[FieldProperty]
    docstring: Optional[str] = ""


@define
class EnumValue:
    """encapsulates enumValues from introspection, maps to an Enum member."""

    name: str
    description: str = ""


@define
class GqlEnumDefinition:
    name: str
    members: list[EnumValue]


EnumMap = dict[str, "GqlEnumDefinition"]


class GqlTypeHinter(TypeHinter):
    def __init__(
        self,
        type: Any,  # noqa: A003
        of_type: tuple["GqlTypeHinter", ...] = (),
    ):
        self.type = type
        self.of_type: tuple["GqlTypeHinter", ...] = of_type

    @property
    def is_object_type(self) -> Optional[GqlTypeDefinition]:
        with contextlib.suppress(TypeError):
            if issubclass(self.type, AntiForwardRef):
                ret = self.type.resolve()
                if isinstance(ret, GqlTypeDefinition):
                    return ret

    @property
    def is_model(self) -> Optional[GqlTypeDefinition]:
        if self.is_list():
            # enums are not supported in lists yet (valid graphql spec though)
            return self.of_type[0].is_object_type

    @property
    def is_enum(self) -> Optional["GqlEnumDefinition"]:
        with contextlib.suppress(TypeError):
            if issubclass(self.type, AntiForwardRef):
                if isinstance(self.type.resolve(), GqlEnumDefinition):
                    return self.type.resolve()

    @property
    def is_builtin_scalar(self) -> Optional[BuiltinScalar]:
        if isinstance(self.type, BuiltinScalar):
            return self.type

    def is_custom_scalar(self, scalars: CustomScalarMap) -> Optional[Type[BaseCustomScalar]]:
        if self.type in scalars.values():
            return self.type

    def annotation(self, scalars: CustomScalarMap) -> str:
        """
        :returns: Annotation of the field based on the real type,
        meaning that the private attribute would be of that type.
        this goes for init and the property setter.
        """
        # int, str, float etc...
        if builtin_scalar := self.is_builtin_scalar:
            return builtin_scalar.tp.__name__

        if scalar := self.is_custom_scalar(scalars):
            return f"SCALARS.{scalar.__name__}"
        if gql_enum := self.is_enum:
            return gql_enum.name
        # handle Optional, Union, List etc...
        # removing redundant prefixes.
        if model_of := self.is_model:
            return f"{QGraphQListModel.__name__}[{model_of.name}]"
        if object_def := self.is_object_type:
            return object_def.name
        if self.is_union():
            return "Union[" + ",".join((th.annotation(scalars) for th in self.of_type)) + "]"
        raise NotImplementedError  # pragma no cover

    def as_annotation(self, object_map=None):  # pragma: no cover
        raise NotImplementedError("not safe to call on this type")



from abc import ABC, abstractmethod
from typing import Any, Generic, Optional, TypeVar

from attr import define

T = TypeVar("T")

__all__ = ["BaseCustomScalar", "BuiltinScalars"]


class BaseCustomScalar(Generic[T], ABC):
    """Class to extend by user defined scalars."""

    __slots__ = "_value"
    GRAPHQL_NAME: str
    """The *real* GraphQL name of the scalar (used by the codegen inspection
    pipeline)."""
    DEFAULT_VALUE: T
    """A place holder graphql query returned null or the field wasn't queried.

    can be used by `from_graphql()`
    """

    def __init__(self, v: Optional[T] = None):
        if not v:
            self._value = self.DEFAULT_VALUE
        else:
            self._value = v

    @abstractmethod
    def from_graphql(cls, v: Optional[Any] = None) -> "BaseCustomScalar":
        """Deserializes data fetched from graphql, This is useful when you want
        to set a first-of value that will later be used by `to_qt()`.

        **must be overridden**!
        """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def to_qt(self) -> Any:
        """Will be used by the property getter, This is the official value that
        Qt should "understand".

        **must be overridden**!
        """
        raise NotImplementedError  # pragma: no cover


@define
class BuiltinScalar:
    tp: type
    default_value: Any
    graphql_name: str


class _BuiltinScalars:
    INT = BuiltinScalar(int, 0, graphql_name="Int")
    FLOAT = BuiltinScalar(float, 0.0, graphql_name="Float")
    STRING = BuiltinScalar(str, " - ", graphql_name="String")
    ID = BuiltinScalar(str, "9b2a0828-880d-4023-9909-de067984523c", graphql_name="ID")
    BOOLEAN = BuiltinScalar(bool, False, graphql_name="Boolean")
    UUID = BuiltinScalar(str, "9b2a0828-880d-4023-9909-de067984523c", graphql_name="UUID")

    @classmethod
    def __iter__(cls):
        for name, member in cls.__dict__.items():
            if name.isupper():
                yield member

    def by_graphql_name(self, name: str) -> Optional[BuiltinScalar]:
        for scalar in self:
            if scalar.graphql_name == name:
                return scalar

    def by_python_type(self, tp: type) -> Optional[BuiltinScalar]:
        for scalar in self:
            if scalar.tp is tp:
                return scalar

    def keys(self) -> list[str]:
        return [scalar.graphql_name for scalar in self]


BuiltinScalars = _BuiltinScalars()



from __future__ import annotations
from PySide6.QtCore import Property, Signal, QObject, QEnum
from PySide6.QtQml import QmlElement
from enum import Enum, auto
from typing import Optional, Union
from qtgql import qproperty
from qtgql.codegen.py.bases import QGraphQListModel
from qtgql.codegen.py.config import QtGqlConfig



{% for dep in context.dependencies %}
{{dep}}{% endfor %}


QML_IMPORT_NAME = "QtGql"
QML_IMPORT_MAJOR_VERSION = 1


{% for enum in context.enums %}
class {{enum.name}}(Enum):
    {% for member in enum.members %}
    {{member.name}} = auto()
    """{{member.description}}"""{% endfor %}

{% endfor %}

{% if context.enums %}
@QmlElement
class Enums(QObject):
    {% for enum in context.enums %}
    QEnum({{enum.name}})
    {% endfor %}
{% endif %}

class SCALARS:
    {% for scalar in context.custom_scalars %}
    {{scalar}} = {{scalar}}{% endfor %}

{% for type in context.types %}
class {{ type.name }}({{context.base_object_name}}):
    """{{  type.docstring  }}"""

    DEFAULT_INIT_DICT = dict({% for f in type.fields %}
        {{f.name}}=None,{% endfor %}
    )
    def __init__(self, parent: QObject = None, {% for f in type.fields %} {{f.name}}: {{f.annotation}} = None, {% endfor %}):
        super().__init__(parent){% for f in type.fields %}
        self.{{  f.private_name  }} = {{f.name}} if {{f.name}} else {{f.default_value}}{% endfor %}

    @classmethod
    def from_dict(cls, parent,  data: dict) -> {{type.name}}:
        init_dict = cls.DEFAULT_INIT_DICT.copy()
        {% for f in type.fields %}
        if {{f.name}} := data.get('{{f.name}}', None):
            init_dict['{{f.name}}'] = {{f.deserializer}}{% endfor %}
        return cls(
            parent=parent,
            **init_dict
        )

    {% for f in type.fields %}
    {{ f.signal_name }} = Signal()

    def {{ f.setter_name }}(self, v: {{  f.annotation  }}) -> None:
        self.{{  f.private_name  }} = v
        self.{{  f.name  }}Changed.emit()

    @qproperty(type={{f.property_type}}, fset={{ f.setter_name }}, notify={{f.signal_name}})
    def {{ f.name }}(self) -> {{ f.fget_annotation }}:
        {{f.fget}}
    {% endfor %}


{% endfor %}



import os

from .qproperty import qproperty
from .slot import slot

if not os.environ.get("QT_API", None):  # pragma: no cover
    os.environ["QT_API"] = "pyside6"


__all__ = ["itemsystem", "slot", "autoproperty", "qproperty"]



from mypy.plugin import Plugin
from mypy.plugins.attrs import attr_define_makers

attr_define_makers.add("qtgql.itemsystem._core.define_roles")
attr_define_makers.add("qtgql.autoproperty.define_properties")


class MyPlugin(Plugin):
    # Our plugin does nothing, but it has to exist so this file gets loaded.
    pass


def plugin(version):
    return MyPlugin



from __future__ import annotations

from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union

import graphql

from qtgql.codegen.py.compiler import TemplateContext
from qtgql.codegen.py.objecttype import (
    EnumMap,
    EnumValue,
    FieldProperty,
    GqlEnumDefinition,
    GqlTypeDefinition,
    GqlTypeHinter,
    Kinds,
)
from qtgql.codegen.py.scalars import BuiltinScalars
from qtgql.codegen.utils import anti_forward_ref

if TYPE_CHECKING:  # pragma: no cover
    from qtgql.codegen.py.config import QtGqlConfig

introspection_query = graphql.get_introspection_query(descriptions=True)


class SchemaEvaluator:
    def __init__(self, introspection: dict, config: QtGqlConfig):
        self.template = config.template_class
        self.introspection = introspection
        self.config = config
        self._generated_types: dict[str, GqlTypeDefinition] = {}
        self._generated_enums: EnumMap = {}
        self._evaluate()

    @cached_property
    def unions(self) -> list[dict]:
        return [
            t for t in self.introspection["__schema"]["types"] if Kinds(t["kind"]) == Kinds.UNION
        ]

    def get_possible_types_for_union(self, name: str) -> list[dict]:
        for union in self.unions:
            if union["name"] == name:
                return union["possibleTypes"]
        raise ValueError(f"Union for {name} was not found")  # pragma: no cover

    def evaluate_field_type(self, t: dict) -> GqlTypeHinter:
        kind = Kinds(t["kind"])

        if kind == Kinds.NON_NULL:
            # There are no optionals in qtgql, we use default values.
            # By default, everything in graphql is optional,
            # so NON_NULL doesn't really make a difference,
            t = t["ofType"]

        of_type = t["ofType"]
        name = t["name"]
        kind = Kinds(t["kind"])
        ret: Optional[GqlTypeHinter] = None

        if kind == Kinds.LIST:
            ret = GqlTypeHinter(type=list, of_type=(self.evaluate_field_type(of_type),))
        elif kind == Kinds.SCALAR:
            if builtin_scalar := BuiltinScalars.by_graphql_name(name):
                ret = GqlTypeHinter(type=builtin_scalar)
            else:
                ret = GqlTypeHinter(type=self.config.custom_scalars[name])
        elif kind is Kinds.ENUM:
            ret = GqlTypeHinter(type=anti_forward_ref(name=name, type_map=self._generated_enums))
        elif kind is Kinds.OBJECT:
            ret = GqlTypeHinter(type=anti_forward_ref(name=name, type_map=self._generated_types))
        elif kind == Kinds.UNION:
            ret = GqlTypeHinter(
                type=Union,
                of_type=tuple(
                    GqlTypeHinter(
                        type=anti_forward_ref(name=possible["name"], type_map=self._generated_types)
                    )
                    for possible in self.get_possible_types_for_union(name)
                ),
            )
        if ret:
            return ret

        raise NotImplementedError(f"kind {kind} not supported yet")  # pragma: no cover

    def evaluate_field(self, field: dict) -> FieldProperty:
        """we don't really know what is the field type just it's name."""
        return FieldProperty(
            type=self.evaluate_field_type(field["type"]),
            name=field["name"],
            type_map=self._generated_types,
            scalars=self.config.custom_scalars,
            description=field["description"],
            enums=self._generated_enums,
        )

    def evaluate_object_type(self, type_: dict) -> Optional[GqlTypeDefinition]:
        # scalars are swallowed here.
        t_name: str = type_["name"]
        if t_name.startswith("__"):
            return None
        if evaluated := self._generated_types.get(t_name, None):
            return evaluated

        concrete = GqlTypeDefinition(
            kind=Kinds.OBJECT,
            name=t_name,
            docstring=type_["description"],
            fields=[self.evaluate_field(f) for f in type_["fields"]],
        )
        return concrete

    def _evaluate_enum(self, enum: dict) -> Optional[GqlEnumDefinition]:
        name: str = enum["name"]
        if name.startswith("__"):
            return None
        if self._generated_enums.get(name, None):
            return None

        return GqlEnumDefinition(
            name=name,
            members=[
                EnumValue(name=val["name"], description=val["description"])
                for val in enum["enumValues"]
            ],
        )

    def _evaluate(self) -> None:
        types = self.introspection["__schema"]["types"]
        for tp in types:
            kind = Kinds(tp["kind"])
            if kind is Kinds.SCALAR:
                continue
            elif kind is Kinds.OBJECT:
                if object_type := self.evaluate_object_type(tp):
                    self._generated_types[object_type.name] = object_type
            elif kind is Kinds.ENUM:
                if enum := self._evaluate_enum(tp):
                    self._generated_enums[enum.name] = enum

    def generate(self) -> str:
        """:return: The generated schema module as a string."""
        return self.template(
            TemplateContext(
                enums=list(self._generated_enums.values()),
                types=[
                    t
                    for name, t in self._generated_types.items()
                    if name not in BuiltinScalars.keys()
                ],
                config=self.config,
            )
        )

    @classmethod
    def from_dict(cls, introspection: dict, config: QtGqlConfig) -> SchemaEvaluator:
        evaluator = SchemaEvaluator(introspection, config=config)
        return evaluator

    def dump(self, file: Path):
        """:param file: Path to the file the codegen would dump to."""
        with open(file, "w") as fh:
            fh.write(self.generate())



import itertools
from typing import Any

from PySide6 import QtCore as qtc

from qtgql.typingref import TypeHinter


def get_combos(stripped_optionals: list[TypeHinter], concretes: list[TypeHinter]):
    required_annotations = tuple([th.type for th in concretes])
    combos = []
    for i in range(len(stripped_optionals) + 1):
        for subset in itertools.combinations([th.type for th in stripped_optionals], i):
            combos.append(subset + required_annotations)
    return combos


def slot(func):
    def wrapper(func) -> func:
        anots: dict = func.__annotations__
        return_ = anots.pop("return", None)
        if return_ is Any:
            return_ = "QVariant"
        args = [TypeHinter.from_annotations(th) for th in anots.values()]
        stripped_optionals = [th.of_type[0] for th in args if th.is_optional()]
        concretes = [th for th in args if not th.is_optional()]

        if stripped_optionals:
            combos = get_combos(stripped_optionals, concretes)
            for combo in combos:
                func = qtc.Slot(*combo, result=return_)(func)
            return func

        return qtc.Slot(*[th.type for th in concretes], result=return_)(func)

    return wrapper(func)



import json
import logging
import typing
import uuid
from abc import abstractmethod
from collections import deque
from typing import Any, Optional

from attrs import define, field
from PySide6 import QtCore as qtc
from PySide6 import QtNetwork as qtn
from PySide6 import QtWebSockets as qtws

from qtgql import slot
from qtgql.gqltransport.core import EncodeAble, GqlEncoder, QueryPayload, T
from qtgql.typingref import UNSET

logger = logging.getLogger(__name__)

__all__ = ["HandlerProto", "GqlWsTransportClient"]


class PROTOCOL:
    """The WebSocket sub-protocol for this specification is: graphql-transport-
    ws.

    https://github.com/enisdenjo/graphql-ws/blob/master/PROTOCOL.md
    """

    CONNECTION_INIT = "connection_init"
    CONNECTION_ACK = "connection_ack"
    NEXT = "next"
    ERROR = "error"
    COMPLETE = "complete"
    PING = "ping"
    PONG = "pong"
    SUBSCRIBE = "subscribe"


@define(kw_only=True)
class BaseGqlWsTransportMessage(EncodeAble):
    type: str
    payload: Optional[Any] = UNSET


@define(kw_only=True)
class GqlClientMessage(typing.Generic[T], BaseGqlWsTransportMessage):
    id: str = field(factory=lambda: uuid.uuid4().hex)
    type: str = PROTOCOL.SUBSCRIBE
    payload: QueryPayload[T] = UNSET

    @classmethod
    def from_query(cls, query: str) -> "GqlClientMessage":
        return cls(payload=QueryPayload(query=query))


@define(kw_only=True)
class SubscribeResponseMessage(BaseGqlWsTransportMessage):
    id: str


class HandlerProto(typing.Protocol):  # pragma: no cover
    message: GqlClientMessage

    @abstractmethod
    def on_data(self, message: dict) -> None:
        raise NotImplementedError

    @abstractmethod
    def on_error(self, message: list[dict[str, Any]]) -> None:
        raise NotImplementedError

    @abstractmethod
    def on_completed(self) -> None:
        raise NotImplementedError


class MESSAGES:
    CONNECTION_INIT = BaseGqlWsTransportMessage(type=PROTOCOL.CONNECTION_INIT)
    PING = BaseGqlWsTransportMessage(type=PROTOCOL.PING)
    PONG = BaseGqlWsTransportMessage(type=PROTOCOL.PONG)


class GqlWsTransportClient(qtws.QWebSocket):
    SUB_PROTOCOL = "graphql-transport-ws"
    textMessageReceived: qtc.Signal
    connected: qtc.Signal
    disconnected: qtc.Signal
    error: qtc.Signal  # type: ignore

    def __init__(
        self,
        *,
        url: str,
        parent=None,
        ping_interval: int = 50000,
        ping_timeout: int = 1000,
        auto_reconnect: bool = False,
        reconnect_timeout: int = 5000,
        headers: Optional[dict[bytes, bytes]] = None,
    ):
        super().__init__(parent=parent)
        self._ping_is_valid = True
        self._connection_ack = False
        self.reconnect_timer = qtc.QTimer(self)
        if auto_reconnect:
            self.reconnect_timer.setInterval(reconnect_timeout)
            self.reconnect_timer.timeout.connect(self.on_reconnect_timeout)  # type: ignore

        self.url: qtc.QUrl = qtc.QUrl(url)
        self.handlers: dict[str, HandlerProto] = {}
        self.pending_messages: deque[GqlClientMessage] = deque()
        self.ping_timer = qtc.QTimer(self)
        self.ping_timer.setInterval(ping_interval)
        self.ping_timer.timeout.connect(self._on_ping_timeout)  # type: ignore
        self.ping_tester_timer = qtc.QTimer(self)
        self.ping_tester_timer.setInterval(ping_timeout)
        self.ping_tester_timer.timeout.connect(self._on_ping_tester_timeout)  # type: ignore
        self.ws_options = qtws.QWebSocketHandshakeOptions()
        self.ws_options.setSubprotocols(
            [
                self.SUB_PROTOCOL,
            ]
        )
        self.textMessageReceived.connect(self.on_text_message)
        self.connected.connect(self._on_connected)
        self.disconnected.connect(self.on_disconnected)
        self.error.connect(self.on_error)
        req = qtn.QNetworkRequest(self.url)
        req.setUrl(self.url)
        if headers:
            for name, value in headers.items():
                req.setRawHeader(name, value)
        self._init_connection(req)

    def _init_connection(self, request: qtn.QNetworkRequest) -> None:
        """Instantiate the connection with server."""
        self.open(request, self.ws_options)

    @staticmethod
    def dumps(data: Any) -> str:
        return json.dumps(data, cls=GqlEncoder)

    def gql_is_valid(self) -> bool:
        """return True if the CONNECTION_ACK has been received and ping is
        valid."""
        return self.isValid() and self._connection_ack

    def subscribe(self, handler: HandlerProto) -> None:
        client_id = handler.message.id
        self.handlers[client_id] = handler
        if self.isValid():
            self.run_subscription(handler.message)
        else:
            self.pending_messages.append(handler.message)

    # this might be helpful in the future for handling single operations.
    query = subscribe

    def run_subscription(self, message: GqlClientMessage) -> None:
        self.sendTextMessage(self.dumps(message))

    def _on_connected(self):
        logger.info(
            f"{self.__class__.__name__}: connection established with server {self.url.toString()}!"
        )
        self.sendTextMessage(self.dumps(MESSAGES.CONNECTION_INIT))
        if self.reconnect_timer.isActive():
            self.reconnect_timer.stop()

    def on_disconnected(self):
        logger.warning(
            f"disconnected from {self.url.toString()} because: {self.closeReason()}; "
            f"close code: {self.closeCode().value}: {self.closeCode().name}"
        )
        self.ping_timer.stop()
        self.ping_tester_timer.stop()
        self.reconnect_timer.start()

    def on_error(self, error: qtn.QAbstractSocket.SocketError):  # pragma: no cover
        logger.warning(f"error occurred in {self.__class__.__name__}: {error}")
        if not self.isValid():
            self.reconnect_timer.start()

    @slot
    def on_reconnect_timeout(self) -> None:
        if not self.isValid():
            self._init_connection(self.request())

    def on_text_message(self, raw: str) -> None:
        data = json.loads(raw)
        if data.get("id", None):
            message = SubscribeResponseMessage(**data)
            m_type = message.type
            if m_type == PROTOCOL.NEXT:
                self._on_gql_next(message)
            elif m_type == PROTOCOL.ERROR:
                self._on_gql_error(message)
            elif m_type == PROTOCOL.COMPLETE:
                self._on_gql_complete(message)
        else:
            # these are id agnostic handlers.
            message = BaseGqlWsTransportMessage(**data)  # type: ignore
            m_type = message.type
            if m_type == PROTOCOL.CONNECTION_ACK:
                self._on_gql_ack()
            elif m_type == PROTOCOL.PONG:
                self._on_gql_pong()
            elif m_type == PROTOCOL.PING:
                self._on_gql_ping()

    def _on_ping_timeout(self):
        self.sendTextMessage(self.dumps(MESSAGES.PING))
        self.ping_tester_timer.start()

    def _on_ping_tester_timeout(self):
        logger.warning(f"pong timeout reached on {self.requestUrl().url()}")
        self.close(qtws.QWebSocketProtocol.CloseCode.CloseCodeReserved1004, "Pong timeout reached")
        self.ping_tester_timer.stop()

    def _on_gql_ack(self) -> None:
        self.sendTextMessage(self.dumps(MESSAGES.PING))
        self.ping_timer.start()
        self.ping_tester_timer.start()
        self._connection_ack = True
        while self.pending_messages:
            self.run_subscription(self.pending_messages.popleft())

    def _on_gql_pong(self) -> None:
        self.ping_tester_timer.stop()

    def _on_gql_ping(self) -> None:
        self.sendTextMessage(self.dumps(MESSAGES.PONG))

    def _on_gql_error(self, message: SubscribeResponseMessage):
        logger.warning("GQL error occurred: %s", message)
        if message.payload:
            self.handlers[message.id].on_error(message.payload)

    def _on_gql_complete(self, message: SubscribeResponseMessage) -> None:
        handler = self.handlers.pop(message.id)
        handler.on_completed()

    def _on_gql_next(self, message: SubscribeResponseMessage) -> None:
        if message.payload:
            self.handlers[message.id].on_data(message.payload["data"])
